// -----------------------------------------------
// Computation of B-Series
//
// Main Program

// -----------------------------------------------
// includes

#include "main.hpp"
#include "tools.hpp"
#include "parc_routines.hpp"
#include "vectorfield.hpp"
#include "tree_structures.hpp"
#include "dag.hpp"
#include "integration.hpp"

#define DIM 3

// -----------------------------------------------
int main(int argc, char* argv[])
{
	// -----------------------------------------------
	// multiprecision
	#ifdef FERI_MPFR
		ScalarType::setDefaultPrecision(MPrecision);
		ScalarType::setDefaultRndMode(ScalarType::RoundNearest);
	#endif

	// output precision
	std::cout.precision(15);

	// -----------------------------------------------
	// max order of the trees = FIRST ARGUMENT
	int max_order = atoi(argv[1]);

	// -----------------------------------------------
	// DAG
	GraphType DAG;

	// file to store the DAG
	std::string dag_file_name = "wDAG/DAG_order_" + boost::lexical_cast<std::string>(max_order) + ".gv";

	// -----------------------------------------------
	// list of trees
	TreeList tree_list;

	#ifdef FERI_DEBUG
		std::cout << "generating the list of rooted trees with max " << max_order << " vertices..." << std::flush;
	#endif

	// generate the list of trees of order at most max_order
	tree_list.generate(max_order);

	#ifdef FERI_DEBUG
		std::cout << " done: " << tree_list.size() << " trees are stored." << std::endl << std::flush;
	#endif

	// -----------------------------------------------
	// IF SECOND ARGUMENT = create
	// we create a DAG and save it to a file
	if (std::strcmp(argv[2], "create") == 0 )
	{

		// -----------------------------------------------
		#ifdef FERI_DEBUG
			std::cout << "generating the DAG from the list..." << std::flush;
		#endif

		// creating the DAG
		generateDAGfromList(DAG, tree_list);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

		// -----------------------------------------------
		#ifdef FERI_DEBUG
			std::cout << "writing the DAG to the file " << dag_file_name << "..." << std::flush;
		#endif

		// writing the DAG
		write_DAG(DAG, dag_file_name);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

	// -----------------------------------------------
	// IF SECOND ARGUMENT = integrate
	// we read the DAG and integrate
	} else if (std::strcmp(argv[2], "integrate") == 0 ) {

		// -----------------------------------------------
		#ifdef FERI_DEBUG
			std::cout << "reading the DAG from the list..." << std::flush;
		#endif

		// reading the DAG
		read_DAG(DAG, dag_file_name);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

			/*
		 			std::pair<VertexIterator, VertexIterator> GRAPHlimits;
					VertexIterator it;
					IndexMap index = get(boost::vertex_index, DAG);
					NameMap name = get(boost::vertex_name, DAG);
					GRAPHlimits = boost::vertices(DAG);
					for (it = GRAPHlimits.first; it != GRAPHlimits.second; ++it)
					std::cout << name[*it] << " " << index[*it] << " " << *it << std::endl;
			*/

		#ifdef FERI_DEBUG
			std::cout << "recreating the order of the vertices in the DAG..." << std::flush;
		#endif

		VertexList vertex_list;

		// putting the vertices in appropriate order into vertex_list
		order_vertexNames(DAG, vertex_list);

		// copying the order from vertex_list to the vertex component
		tree_list.match_vertices(vertex_list);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

		// -----------------------------------------------
		// defining the vectorfield
		typedef VectorField<DIM, DIM> VectorFieldType;
		VectorFieldType F;

		// -----------------------------------------------
		// the point to derivate at;
		VectorType x(DIM);

		// reading initial conditions from the file
		read_from_file("input/initial_conditions.txt", x);

		#ifdef FERI_DEBUG
			std::cout << "We start to integrate from: " << x << std::endl;
		#endif

		// -----------------------------------------------
		// the alpha values
		capd::vectalg::Vector<ScalarType, 0> alpha(tree_list.size());

		// reading the alpha values
		read_from_file("input/alpha.txt", alpha);

		#ifdef FERI_DEBUG
			std::cout << "Alpha: " << alpha << std::endl;
		#endif

		// -----------------------------------------------
		// the integration
		std::cin.ignore(1);

		#ifdef FERI_DEBUG
			std::cout << "Integration has started." << std::endl;
		#endif

		while(true)
		{
			std::cout << x << std::endl;
			std::cin.ignore(1);
			x = integrate(F, x, DAG, tree_list, alpha);
		}

	// -----------------------------------------------
	// IF SECOND ARGUMENT = testDAG
	// we read the DAG and test
	} else if (std::strcmp(argv[2], "testDAG") == 0 ) {

	// -----------------------------------------------
		#ifdef FERI_DEBUG
			std::cout << "reading the DAG from the list..." << std::flush;
		#endif

		// reading the DAG
		read_DAG(DAG, dag_file_name);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

		#ifdef FERI_DEBUG
			std::cout << "recreating the order of the vertices in the DAG..." << std::flush;
		#endif

		VertexList vertex_list;

		// putting the vertices in appropriate order into vertex_list
		order_vertexNames(DAG, vertex_list);

		// copying the order from vertex_list to the vertex component
		tree_list.match_vertices(vertex_list);

		#ifdef FERI_DEBUG
			std::cout << " done" << std::endl << std::flush;
		#endif

		// -----------------------------------------------
		// defining the vectorfield
		typedef VectorField<DIM, DIM> VectorFieldType;
		VectorFieldType F;
		// -----------------------------------------------
		// the point to derivate at;
		VectorType x(DIM);

		// reading initial conditions from the file
		read_from_file("input/initial_conditions.txt", x);

		#ifdef FERI_DEBUG
			std::cout << "We test at the point: " << x << std::endl;
		#endif

		// -----------------------------------------------
		// the alpha values
		capd::vectalg::Vector<ScalarType, 0> alpha(tree_list.size());

		// reading the alpha values
		read_from_file("input/alpha.txt", alpha);

		#ifdef FERI_DEBUG
			std::cout << "Alpha: " << alpha << std::endl;
		#endif

		// -----------------------------------------------
		// the test

		#ifdef FERI_DEBUG
			std::cout << "Testing has started." << std::endl;
		#endif

		test(F, x, DAG, tree_list, alpha);

	// -----------------------------------------------
	// IF SECOND ARGUMENT = testTAYLOR
	// we give back the result of a Taylor step
	} else if (std::strcmp(argv[2], "testTAYLOR") == 0 ) {

		// -----------------------------------------------
		// the initial condition
		VectorType x(DIM);

		// reading initial conditions from the file
		read_from_file("input/initial_conditions.txt", x);		
		
		// -----------------------------------------------
		// the basic fad taylor
		typedef TestVectorField<ScalarType, DIM> TVtype;
		TVtype tvfield;

		capd::dynsys::BasicFadTaylor<TVtype> btaylor(tvfield, 5, 0.1);
		capd::vectalg::Vector<ScalarType, DIM> tvector;

		for (int i = 1; i <= DIM; ++i)
			tvector(i) = x(i);

		tvector = btaylor(tvector);
		std::cout << "Taylor: " << tvector << std::endl;

	// -----------------------------------------------
	// IF SECOND ARGUMENT = testPARC
	// we give back a partial derivative
	} else if (std::strcmp(argv[2], "testPARC") == 0 ) {

		// -----------------------------------------------
		// defining the vectorfield
		typedef VectorField<DIM, DIM> VectorFieldType;
		VectorFieldType F;

		// -----------------------------------------------
		// the point to derivate at
		VectorType x(DIM);

		// reading initial conditions from the file
		read_from_file("input/initial_conditions.txt", x);
		
		// reading the multiindex i and the seedmatrix
		int sdim;
		read_from_file("test/input/testPARC_dimofS.txt",sdim);
		SeedMatrixType S(DIM, sdim);
		MultiIndexType i(sdim);
		read_from_file("test/input/testPARC_multiindex.txt", i);
		read_from_file("test/input/testPARC_seedmatrix.txt", S);

		std::cout << "We integrate at the point: " << std::endl << x << std::endl;
		std::cout << "With the seedmatrix: " << std::endl << S << std::endl;
		std::cout << "And the multiindex: " << std::endl << i << std::endl;
		std::cout << "The partial derivative is: " << std::endl << parc_deriv(F, x, S, i, ConvertToInt(sum(i))) << std::endl;
	}

}
// ~ end of file ~
