#ifndef FERI_DAG
#define FERI_DAG
// -----------------------------------------------
// Computation of B-Series
//
// Directed Acyclic Graph

// -----------------------------------------------

#include "main.hpp"
#include "tree_structures.hpp"

// -----------------------------------------------
// generate DAG from a tree list
void generateDAGfromList(GraphType &G, TreeList &tree_list)
{
	typedef TreeList::iterator TreeListIterator;
	//TreeListElement list_element(1);

	VertexDescriptor v, w;
	EdgeAccess edge;

	EdgeWeightMap weight_map = get(boost::edge_weight, G);
	int weight;

/*
	// we record the positions of different size regions in the tree list
	int current_size = 0;
	TreeListIterator it = tree_list.end();
	--it;

	std::vector<TreeListIterator> RegionStorage;
	//( ((*it).tree).size() );
	for (it = tree_list.begin(); it != tree_list.end(); ++it) {
		if ( ((*it).tree).size() > current_size ) {
			++current_size;
			RegionStorage(current_size) = it;
		}
	}
*/
	// we initialize the DAG
	G.clear();

	#ifdef FERI_DEBUG
		long int i = 0;
		long int size = tree_list.size();
		int percentage = 10;
		std::cout << std::endl;
	#endif

	// we go through our storage of trees
	for (TreeListIterator it = tree_list.begin(); it != tree_list.end(); ++it)
	{

		#ifdef FERI_DEBUG
			++i;
			if (100.0 * i / size > percentage) {
				std::cout << percentage << "% ready..." << std::endl;
				percentage = percentage + 10;
			}
		#endif

		// we add a new vertex for the root of the current tree
		v = boost::add_vertex(G);
		(*it).vertex = v;

		// we connect it to the roots of its children with appropriate weight (since they are of smaller order, they are already processed)
		for (int i = 1; i <= (*it).tree.numberOfChildren(); ++i) {

			// the current children
			w = (*tree_list.find( (*it).tree.child(i) )).vertex;
			//list_element.tree = (*it).tree.child(i);
			//w = (*std::find(tree_list.begin(), tree_list.end(), list_element)).vertex;

			// the edge between the root and its children
			edge = boost::edge(v, w, G);

			// there is an edge already, we increase its weight
			if (edge.second) {

				weight = get(weight_map, edge.first);
				put(weight_map, edge.first, weight + 1);

			// there was no such edge yet
			} else {

				// add with weight 1
				edge = boost::add_edge(v, w, G);
				put(weight_map, edge.first, 1);

			}

			//

		}

	}

	#ifdef FERI_DEBUG
		std::cout << percentage << "% ready...";
	#endif

}

// -----------------------------------------------
// read in DAG from a file
void read_DAG(GraphType &G, const std::string& filename)
{
	std::ifstream in(filename.c_str());

	boost::dynamic_properties dp;
	dp.property("weight", get(boost::edge_weight, G));
	dp.property("id", get(boost::vertex_name, G));

	boost::read_graphviz(in, G, dp, "id");

	in.close();
}

// -----------------------------------------------
// write DAG to a file
void write_DAG(GraphType &G, const std::string& filename)
{
	std::ofstream out(filename.c_str());

	boost::dynamic_properties dp;
	dp.property("weight", get(boost::edge_weight, G));
	dp.property("index", get(boost::vertex_index, G));

	boost::write_graphviz_dp(out, G, dp, std::string("index"));

	out.close();
}

// -----------------------------------------------
// post-ordering of a graph
void rec_post_order(GraphType &G, VertexDescriptor v, VertexList &VList)
{
	std::pair<OutEdgeIterator, OutEdgeIterator> Edges;
	OutEdgeIterator e;

	Edges = boost::out_edges(v, G);
	e = Edges.first;

	while ( e != Edges.second ) {

		rec_post_order(G, boost::target(*e, G), VList);
		++e;

	}

	VList.push_back(v);

}

void post_order(GraphType &G, VertexList &VList)
{
	rec_post_order(G, *(boost::vertices(G).first), VList);
}

// -----------------------------------------------
// evaluate the quantity represented by the tree
template<typename MapType, typename VType>
void evaluate_graph(MapType &F, const VType &x, GraphType &G, TreeList tree_list)
{

	for (TreeList::iterator it = tree_list.begin(); it != tree_list.end(); ++it)
		evaluate_vertex(F, x, G, (*it).vertex);

}

// evaluating a node
template<typename MapType, typename VType>
void evaluate_vertex(MapType &F, const VType &x, GraphType &G, VertexDescriptor v)
{
	#ifdef FERI_DEBUG_DAG
		NameMap name = get(boost::vertex_name, G);

		std::cout << "at node: " << name[v] << std::endl << std::flush;
	#endif

	ColorMap color = get(boost::vertex_color, G);
	EdgeWeightMap weight = get(boost::edge_weight, G);

	int t_degree = boost::out_degree(v, G);

	int m = F.dimension().second;
	VectorType temp(m);

	// function value
	if (t_degree == 0) {

		F(x, temp);
		put(color, v, temp);

		#ifdef FERI_DEBUG_DAG
			std::cout << "DAG >> function value: " << temp << std::endl << std::flush;
		#endif

	// induction rule
	} else {

		// defining the seed matrix S and the multi index i
		SeedMatrixType S(m, t_degree);
		MultiIndexType i(t_degree);

		// filling up i and S
		std::pair<OutEdgeIterator, OutEdgeIterator> Edges;
		OutEdgeIterator e;

		Edges = boost::out_edges(v, G);
		e = Edges.first;

		int j = 1;
		while ( e != Edges.second ) {

			i(j) = get(weight, *e);
			for (int k = 1; k <= m; ++k)
				S(k, j) = get(color, boost::target(*e, G))(k);
			++j;
			++e;

		}

		#ifdef FERI_DEBUG_DAG
			std::cout << "DAG >> " << sum(i) << "th derivative" << std::endl << std::flush;
		#endif

		// calculating the derivative
		put(color, v, parc_deriv(F, x, S, i, ConvertToInt(sum(i))));

	}

}

// comparison for sorting algorithm
// -----------------------------------------------
// order the list of vertices by their names
class comp_vertex_operator {
public:

	// the operator
	bool operator()(VertexDescriptor a, VertexDescriptor b)
	{
		NameMap name = get(boost::vertex_name, *G);
		return (atoi( name[a].c_str() ) < atoi( name[b].c_str() ));
	}

	// set the pointer to the graph
	void setGraph(GraphType &graph){
		G = &graph;
	}

	// the pointer to the graph
	GraphType *G;
};

// fill vertex list and sort
	void order_vertexNames(GraphType &G, VertexList &vertex_list)
	{
		std::pair<VertexIterator, VertexIterator> GRAPHlimits;
		GRAPHlimits = boost::vertices(G);
		for (VertexIterator it = GRAPHlimits.first; it != GRAPHlimits.second; ++it)
			vertex_list.push_front(*it);

		comp_vertex_operator comp_vertex;
		comp_vertex.setGraph(G);

		vertex_list.sort(comp_vertex);
	}

// ~ end of file ~
#endif
