#ifndef FERI_PARCROUTINES
#define FERI_PARCROUTINES
// -----------------------------------------------
// Computation of B-Series
//
// Parc Diff Routines

// -----------------------------------------------

#include "main.hpp"

// -----------------------------------------------
// typedefs

// Taylor expansion in Scalars
typedef fadbad::T<ScalarType> TScalar;

// -----------------------------------------------
// conversion to integer
template <typename T>
static inline int ConvertToInt(const T &a)
{
	#ifdef FERI_MPFR
		return capd::multiPrec::toInt(a);
	#else
		return int(a);
	#endif
}

// -----------------------------------------------
// routines for multi-indices

// Sum of the components of a multi-index
template<typename MultiIndex>
ScalarType sum(const MultiIndex &a)
{
	ScalarType result = 0;

	for (int i = 1; i <= a.size(); ++i)
		result += a(i);

	return result;
}

// -----------------------------------------------
// routines for gamma(i, j)

// n over k, where k is positive integer
ScalarType binomial(const ScalarType &n, const ScalarType &k)
{
	ScalarType result = 1.0;

	for (int i = ConvertToInt(k); i > 0; --i)
		result *= (n - i + 1) / ScalarType(i);

	return result;
}

// i over j, where they are multi-indices, j consists of non-negative integers
template<typename MultiIndex>
ScalarType multi_binomial(const MultiIndex &i, const MultiIndex &j)
{
	ScalarType result = 1.0;

	for (int temp = 1; temp <= j.size(); ++temp)
		result = result * binomial(i(temp), j(temp));

	return result;
}

// inner routine of getting gamma(i,j)
template<typename MultiIndex>
void rec_gamma(const MultiIndex &i, const MultiIndex &j, const int &d, MultiIndex k, int pos, ScalarType &result)
{
	if (pos == i.size()) {
		for (; k(pos) <= i(pos); k(pos) = k(pos) + 1) {
			if (sum(k) != 0)
				result += pow(-1, sum(i - k)) * multi_binomial(i, k) * multi_binomial(d * k / ScalarType(sum(k)), j) * pow(sum(k) / ScalarType(d), sum(i));
		}
	} else {
		for (; k(pos) <= i(pos); k(pos) = k(pos) + 1)
			rec_gamma(i, j, d, k, pos + 1, result);
	}
}

// getting gamma(i,j)
template<typename MultiIndex>
ScalarType gamma(const MultiIndex &i, const MultiIndex &j, const int &d)
{
	ScalarType result = 0.0;
	MultiIndex k(i.size());

	for (int m = 1; m <= k.size(); ++m)
		k(m) = 0;

	rec_gamma(i, j, d, k, 1, result);

	return result;
}

// -----------------------------------------------
// routines for getting Taylor coefficients

// getting the appropriate Taylor coefficient
template<typename MapType, typename VType, typename V2Type>
VectorType TaylorCoeff(MapType &F, const VType &x, const V2Type &s, const ScalarType &i)
{
	VectorType result(F.dimension().second);

	F.evaluate_from(x, s, ConvertToInt(i), result);

	return result;
}

// -----------------------------------------------
// routines to obtain a given partial derivative

// inner routine of summing over ¦j¦=d
template<typename MapType, typename VType, typename MultiIndex, typename SeedMatrix>
void rec_parc_deriv(MapType &F, const VType &x, const SeedMatrix &S, const MultiIndex &i, MultiIndex j, const int &d, int pos, int d_rest, VectorType &result)
{

	if (pos == j.size()) {
			j(pos) = d_rest;

			#ifdef FERI_DEBUG_PARC
				std::cout << "PARC >> GAMMA( " << i << ", " << j << "): " << gamma(i, j , d) << std::flush;
				std::cout << " TaylorCoeff: " << TaylorCoeff(F, x, S * j, sum(i)) << std::endl << std::flush;
			#endif

			result += gamma(i, j, d) * TaylorCoeff(F, x, S * j, sum(i));
	} else {
		for (; j(pos) <= d_rest; j(pos) = j(pos) + 1)
			rec_parc_deriv(F, x, S, i, j, d, pos + 1, ConvertToInt(d_rest - j(pos)), result);
	}

}

// returning the comp component of D^|i|/(S.^i) F(x)
// summing over ¦j¦=d
template<typename MapType, typename VType, typename MultiIndex, typename SeedMatrix>
VectorType parc_deriv(MapType &F, const VType &x, const SeedMatrix &S, const MultiIndex &i, const int &d)
{
	#ifdef FERI_DEBUG_PARC
		std::cout << "PARC >> Evaluating parc derivative from point " << x << " with d: " << d << " and index " << i << std::endl << "directions: " << std::endl << S << std::endl << std::flush;
	#endif

	MultiIndex j(i.size());
	VectorType result(F.dimension().second);

	for (int m = 1; m <= j.size(); ++m)
		j(m) = 0;

	rec_parc_deriv(F, x, S, i, j, d, 1, d, result);

	#ifdef FERI_DEBUG_PARC
		std::cout << "PARC >> The partial derivative is: " << result << std::endl << std::flush;
	#endif

	return result;
}

// returning a partial derivative
template<typename MapType, typename VType>
VectorType partial_derivative(MapType &F, const VType &x, const IndexType &parc)
{
	int n, temp, numDirections;

	temp = 0;
	n = F.dimension().first;

	// how many directions we need
	numDirections = 0;
	for (int j = 1; j <= n; ++j)
			if (parc(j) != 0)
				++numDirections;

	// defining the seed matrix S and the multi index i
	SeedMatrixType S(n, numDirections);
	MultiIndexType i(numDirections);

	// filling up i and S
	for (int j = 1; j <= n; ++j) {
		if (parc(j) != 0) {
			++temp;
			i(temp) = parc(j);
			S(j, temp) = 1;
		}
	}

	// calculating the derivative
	return parc_deriv(F, x, S, i, ConvertToInt(sum(i)));
}

// ~ end of file ~
#endif
